!-------------------------------------------------------------------
!
!   this program solves AB block copolymer melt Gyroid structure.
!   3-dimensional SCFT (pseudospectral method + Periodic boundary +
!   + Anderson mixing) is chosen.
!
!-------------------------------------------------------------------
!this module defines parameters defined once and used repeatedly.
!only static parameters are included in this module
module parameters
    implicit none
!   chiN = interaction parameter between A and B monomers,  pi = 3.141592..
    double precision :: chiN
    double precision, parameter :: pi = atan(1.0d0)*4.0d0
!
!   grid parameters
!   II, JJ, KK = number of grids in x, y and z direction
!   NN = number of time steps (in s direction)
!   NNf = number of time steps for A fractio
    integer :: II, JJ, KK, NN, NNf
!   Lx, Ly, Lz = length of the system in each direction (in units of aN^1/2)
!   dx, dy, dz, ds = discrete step sizes
!   f = A fraction (1-f is the B fraction)
    double precision :: Lx, Ly, Lz, dx, dy, dz, ds, f
!   seg = simple integral weight (half contribution at both edges)
    double precision, allocatable :: segx(:),segy(:),segz(:),seg(:,:,:),expfactor(:,:,:)
!   volume = volume of the system
    double precision :: volume
end module parameters
!
!--------------------------------------------------------------------
!   The main program begins here
program three_dim_block_copolymer
!
    use mkl_service
    use parameters
    implicit none
!   iter = number of iteration steps, maxiter = maximum number of iteration steps
    integer :: i, j, k, iter, maxiter
    double precision :: xx,yy,zz,c1,c2
    logical :: has_input
!   parameters related to Anderson mixing
    integer :: nit, nit2, n_anderson, start_anderson, max_anderson, m
!   QQ = total partition function
    double precision :: QQ, energy_chain, energy_field, energy_tot, energy_old
!   error_level = variable to check convergence of the field
!   dot = function performing dot product
!   multidot = function performing multiple dot product
!   mix = dynamically changing mixing parameter
!   min_mix = its minimum value
    double precision :: error_level, old_error_level, dot, multidot
    double precision :: mix, min_mix, temp
!   input(w) and output(wout) fields, xi is temporarily accepted pressures
    double precision, allocatable :: w(:,:,:,:),wout(:,:,:,:), xi(:,:,:)
!   previous field values stored for Anderson mixing
    double precision, allocatable :: wout_hist(:,:,:,:,:),wdiff_hist(:,:,:,:,:)
!   arrays to calculate Anderson mixing
    double precision, allocatable :: u_nm(:,:), v_n(:), a_n(:), wdiffdots(:,:)
!   A and B segment concentration and total segment concentration
    double precision, allocatable :: phia(:,:,:), phib(:,:,:), phitot(:,:,:)
!   strings to name output files
    character (len=5) extension
    character (len=20) label
!
!   has_input = .true. if input file is used
    has_input = .false.
!
    if (has_input) then
        open(10,file='input.txt', status='unknown')
!       read all the stored parameters
        read(10,*) extension, mix, chiN, f, energy_tot, error_level
        read(10,*) NN, II, JJ, KK, Lx, Ly, Lz
!       read fields and concentrations from the file
        allocate(w(1:2,1:II,1:JJ,1:KK),wout(1:2,1:II,1:JJ,1:KK),xi(1:II,1:JJ,1:KK))
        allocate(phia(1:II,1:JJ,1:KK),phib(1:II,1:JJ,1:KK),phitot(1:II,1:JJ,1:KK))
        do i=0,II
            do j=0,JJ
                do k=0,KK
                    read(10,*) w(1,i,j,k), phia(i,j,k), w(2,i,j,k), phib(i,j,k)
                end do
            end do
        end do
        close(10)
        phitot = phia + phib
!
    else
!   if has_input = .false. set up the parameters of the system we solve
!   block copolymer sepcification
        chiN = 20.0d0
        f = 0.36d0
        II = 32
        Lx = 3.88925d0
        JJ = 32
        Ly = 3.88925d00
        KK = 32
        Lz = 3.88925d0
        NN = 25
    end if
!
!   mixing parameter
    mix = 0.2d0
!   minimum mixing parameter
    min_mix = 0.1d0
!   number of iterations before Anderson mixing begins.
!   for now, we assign a large value. it will be determined later.
    start_anderson = 10000
!   max number of previous steps to calculate new field when using Anderson mixing
    max_anderson = 30
    m = max_anderson + 1
!   number of Anderson mixing steps, increases from 1 to max_anderson
    n_anderson = 0
!   iteration must stop before iter = maxiter happens
    maxiter = 1000
!
!   parameter initialization
    call initialize()
!
!   define array sizes for Anderson mixing
    allocate(wout_hist(0:max_anderson,1:2,1:II,1:JJ,1:KK),wdiff_hist(0:max_anderson,1:2,1:II,1:JJ,1:KK))
    allocate(u_nm(1:max_anderson,1:max_anderson),v_n(1:max_anderson),a_n(1:max_anderson),wdiffdots(0:max_anderson,0:max_anderson))
!
!   create initial field
    if(.not.has_input) then
        allocate(w(1:2,1:II,1:JJ,1:KK),wout(1:2,1:II,1:JJ,1:KK))
        allocate(phia(1:II,1:JJ,1:KK),phib(1:II,1:JJ,1:KK),phitot(1:II,1:JJ,1:KK))
        do i=1,II
            xx = i*2.0d0*pi/II
            do j=1,JJ
                yy = j*2.0d0*pi/JJ
                do k=1,KK
                    zz = k*2.0d0*pi/KK
                    c1 = sqrt(8.0d0/3.0d0)*(cos(xx)*sin(yy)*sin(2.0d0*zz)+&
                        cos(yy)*sin(zz)*sin(2.0d0*xx)+cos(zz)*sin(xx)*sin(2.0d0*yy))
                    c2 = sqrt(4.0d0/3.0d0)*(cos(2.0d0*xx)*cos(2.0d0*yy)+&
                        cos(2.0d0*yy)*cos(2.0d0*zz)+cos(2.0d0*zz)*cos(2.0d0*xx))
!
                    w(1,i,j,k) = -0.364d0*c1+0.133d0*c2
                    w(2,i,j,k) = 0.302d0*c1-0.106*c2
                end do
            end do
        end do
!   keep the level of field value (in order to prevent drifting)
!        call zerofield(w)
    end if
!   assign large initial value for the energy and error
    energy_tot =100.0d0
    error_level = 100.0d0
!
!   iteration begins here
    do iter=1,maxiter
!       for the given fields w_a and w_b, find the segment concentration and Q
        call findphi(exp(-w(1,:,:,:)*ds*0.5d0),exp(-(w(2,:,:,:))*ds*0.5d0),phia,phib,QQ)
        phitot = phia + phib
!       calculate individual energy contribution
        energy_chain = -log(QQ/volume)
        energy_field = (chiN*dot(phia,phib) - dot(w(1,:,:,:),phia) - dot(w(2,:,:,:),phib))/volume
!       calculate the total energy per chain
        energy_old = energy_tot
        energy_tot = energy_chain + energy_field
!       calculate pressure field for the new field calculation
        xi = 0.5d0*(w(1,:,:,:)+w(2,:,:,:)-chiN)
!       calculate output fields
        wout(1,:,:,:) = chiN*phib + xi
        wout(2,:,:,:) = chiN*phia + xi
        call zerofield(wout)
!       error_level measures the "relative distance" between the input and output
!       fields
        old_error_level = error_level
        error_level = sqrt(multidot(wout-w,wout-w)/multidot(w,w))
!       print iteration # and error levels
        write(*,'(I6,F8.3,2F13.9)') iter, mix, energy_tot, error_level
!       check the condition to start Anderson mixing
        if(error_level < 1.0d-2) start_anderson = min(start_anderson,iter)
!       check the condition to end the iteration
        if(error_level < 1.0d-7) exit
!       check the condition to apply the simple mixing method
        if(iter <= start_anderson) then
!           just before the Anderson mixing start accumulating past fields
            if(iter == start_anderson) then
                wout_hist(mod(iter,m),:,:,:,:) = wout
                wdiff_hist(mod(iter,m),:,:,:,:) = wout - w
                wdiffdots(mod(iter,m),mod(iter,m)) &
                    = multidot(wdiff_hist(mod(iter,m),:,:,:,:),wdiff_hist(mod(iter,m),:,:,:,:))
            end if
!           dynamically change mixing parameter
            if(old_error_level < error_level) then
                mix = max(mix*0.7d0,min_mix)
            else
                mix = mix*1.01d0
            end if
!           calculate the new field using simple mixing
            w = (1.0d0-mix)*w + mix*wout
!
        else
!           Anderson mixing begins here
!           number of previous steps to use for Anderson mixing
            n_anderson = min(max_anderson,iter-start_anderson)
!           store the input and output field (the memory is used in a periodic way)
            wout_hist(mod(iter,m),:,:,:,:) = wout
            wdiff_hist(mod(iter,m),:,:,:,:) = wout - w
!           evaluate wdiff dot products for calculating Unm and Vn in Thompson's paper
            wdiffdots(mod(iter,m),mod(iter,m)) &
                = multidot(wdiff_hist(mod(iter,m),:,:,:,:),wdiff_hist(mod(iter,m),:,:,:,:))
            do nit=1,n_anderson
                wdiffdots(mod(iter,m),mod(iter-nit,m)) &
                    = multidot(wdiff_hist(mod(iter,m),:,:,:,:),wdiff_hist(mod(iter-nit,m),:,:,:,:))
                wdiffdots(mod(iter-nit,m),mod(iter,m)) = wdiffdots(mod(iter,m),mod(iter-nit,m))
            end do
!           calculate Unm and Vn
            do nit=1,n_anderson
                v_n(nit) =  wdiffdots(mod(iter,m),mod(iter,m)) &
                    - wdiffdots(mod(iter,m),mod(iter-nit,m))
                do nit2=1,n_anderson
                    u_nm(nit,nit2) =  wdiffdots(mod(iter,m),mod(iter,m)) &
                        - wdiffdots(mod(iter,m),mod(iter-nit,m))&
                        - wdiffdots(mod(iter-nit2,m),mod(iter,m))&
                        + wdiffdots(mod(iter-nit,m),mod(iter-nit2,m))
                end do
            end do
!
            call find_an(u_nm(1:n_anderson,1:n_anderson),v_n(1:n_anderson),a_n(1:n_anderson),n_anderson)
!           calculate the new field using Anderson mixing
            w = wout_hist(mod(iter,m),:,:,:,:)
            do nit=1,n_anderson
                w = w + a_n(nit)*(wout_hist(mod(iter-nit,m),:,:,:,:)-wout_hist(mod(iter,m),:,:,:,:))
            end do
        end if
!
!       create intermediate output file at every 100 steps
        if(mod(iter,100) == 0) then
            write (extension,'(I0.4)') iter
            label = 'output.' // extension
!
            open(20,file=label,status='unknown')
            write(20,'(I8,3F8.3,2F13.9)') iter, mix, chiN, f, energy_tot, error_level
            write(20,'(4I5,3F8.3)') NN, II, JJ, KK, Lx, Ly, Lz
            write(20,*) " "
            do i=1,II
                do j=1,JJ
                    do k=1,KK
                        write(20,'(4F12.6)') w(1,i,j,k), phia(i,j,k), w(2,i,j,k), phib(i,j,k)
                    end do
                end do
            end do
            close(20)
        end if
!   the main loop ends here
    end do
!
!   write the final output
    open(30,file='output.txt',status='unknown')
    write(30,'(I8,3F8.3,2F13.9)') iter, mix, chiN, f, energy_tot, error_level
    write(30,'(4I5,3F8.3)') NN, II, JJ, KK, Lx, Ly, Lz
    write(30,*) " "
    do i=1,II
        do j=1,JJ
            do k=1,KK
                write(30,'(4F12.6)') w(1,i,j,k), phia(i,j,k), w(2,i,j,k), phib(i,j,k)
            end do
        end do
    end do
    close(30)
!
end program three_dim_block_copolymer
!
!--------------------------------------------------------------------
!this subroutine initializes repeatedly used parameters
subroutine initialize()
!
    use parameters
    implicit none
    integer :: i, j, k, jtemp, ktemp
    real(kind=8) :: xfactor, yfactor, zfactor
!
!   s step number for A fraction
    NNf = NN*f
    if( abs(NNf - NN*f) > 1d-8) then
        write(*,*) "NN*f is not an integer."
        stop
    end if
!   grid sizes in x, y, z and s direction
    dx = Lx/II
    dy = Ly/JJ
    dz = Lz/KK
    ds = 1.0d0/NN
    volume = Lx*Ly*Lz
!   define matrix size
    allocate(seg(1:II,1:JJ,1:KK),segx(1:II),segy(1:JJ),segz(1:KK))
    allocate(expfactor(0:II/2,0:JJ-1,0:KK-1))
!
!   weight factor for three-dim integral
    segx(:) = 1.0d0 * dx
    segy(:) = 1.0d0 * dy
    segz(:) = 1.0d0 * dz
!
    do i=1,II
        do j=1,JJ
            do k=1,KK
                seg(i,j,k) = segx(i)*segy(j)*segz(k)
            end do
        end do
    end do
!
! calculate seg and the exponetial factor
    xfactor = -(2*pi/Lx)**2*ds/6.0d0
    yfactor = -(2*pi/Ly)**2*ds/6.0d0
    zfactor = -(2*pi/Lz)**2*ds/6.0d0

    do i=0,II/2
        do j=0,JJ-1
          if(j > JJ/2) then
            jtemp = JJ-j
          else
            jtemp = j
          end if
          do k=0,KK-1
            if(k > KK/2) then
              ktemp = KK-k
            else
              ktemp = k
            end if
            expfactor(i,j,k) = exp(i**2*xfactor+jtemp**2*yfactor+ktemp**2*zfactor)
          end do
        end do
      end do
!
end subroutine initialize
!
!--------------------------------------------------------------------
!This subroutine keeps the level of field
subroutine zerofield(w)
!
    use parameters
    implicit none
    double precision, intent(inout) :: w(1:2,1:II,1:JJ,1:KK)
    double precision :: tot1, tot2, temp
!   each sum of w_a and w_b is set to zero
    tot1 = sum(seg(:,:,:)*w(1,:,:,:))
    w(1,:,:,:) = w(1,:,:,:) - tot1/volume
    tot2 = sum(seg(:,:,:)*w(2,:,:,:))
    w(2,:,:,:) = w(2,:,:,:) - tot2/volume
!
end subroutine zerofield
!
!--------------------------------------------------------------------
!this function calculates three-dim integral g*h
double precision function dot(g,h)
!
    use parameters
    implicit none
    double precision, intent(in) :: g(1:II,1:JJ,1:KK), h(1:II,1:JJ,1:KK)
    dot = sum(seg(:,:,:)*g(:,:,:)*h(:,:,:))
end function dot
!
!--------------------------------------------------------------------
!this function calculates multiple three-dim integrals
double precision function multidot(g,h)
!
    use parameters
    implicit none
    double precision, intent(in) :: g(1:2,1:II,1:JJ,1:KK), h(1:2,1:II,1:JJ,1:KK)
    multidot = sum(seg(:,:,:)*(g(1,:,:,:)*h(1,:,:,:)+g(2,:,:,:)*h(2,:,:,:)))
end function multidot
!
!--------------------------------------------------------------------
!this subroutine finds the solution for the matrix equation (n*n size)
!for Anderson mixing U * A = V
!Gauss elimination method is used in its simplest level
subroutine find_an(u,v,a,n)
    implicit none
    integer, intent(in) :: n
    double precision, intent(inout) :: u(1:n,1:n), v(1:n)
    double precision, intent(out) :: a(1:n)
!
    integer :: i,j,k
    double precision :: factor, tempsum
!   Elimination process
    do i=1,n
        do j=i+1,n
            factor = u(j,i)/u(i,i)
            v(j) = v(j) - v(i)*factor
            do k=i+1,n
                u(j,k) = u(j,k) - u(i,k)*factor
            end do
        end do
    end do
!
!   find the solution
    a(n) = v(n)/u(n,n)
    do i=n-1,1,-1
        tempsum = 0.0d0
        do j=i+1,n
            tempsum = tempsum + u(i,j)*a(j)
        end do
        a(i) = (v(i) - tempsum)/u(i,i)
    end do
!
end subroutine find_an
!
!--------------------------------------------------------------------
!this subroutine calculates the segment concentration
!where expdwa = exp(-wa*ds/2) and expdwb = exp(-wb*ds/2)
subroutine findphi(expdwa,expdwb,phia,phib,QQ)
!
    use parameters
    implicit none
    double precision, intent(in) :: expdwa(1:II,1:JJ,1:KK), expdwb(1:II,1:JJ,1:KK)
    double precision, intent(out) :: phia(1:II,1:JJ,1:KK), phib(1:II,1:JJ,1:KK)
    double precision, intent(out) :: QQ
    double precision :: q1(1:II,1:JJ,1:KK,0:NN), q2(1:II,1:JJ,1:KK)
    double precision :: dot, temp
    integer :: n, i, j, k
!   q1 is q and q2 is qdagger in the note
!   free end initial condition (q1 starts from A end)
    q1(:,:,:,0) = 1.0d0
!
!   diffusion of A chain
    do n=1,NNf
        call pseudo(q1(:,:,:,n-1),q1(:,:,:,n),expdwa)
    end do
!
!   diffusion of B chain
    do n=NNf+1,NN
        call pseudo(q1(:,:,:,n-1),q1(:,:,:,n),expdwb)
    end do
!
!   q2 starts from B end
    q2 = 1.0d0
!   segment concentration from the end. (note the half contribution)
    phib = 0.5d0*q1(:,:,:,NN)
!
!   diffusion of q2 in B part
    do n=NN-1,NNf+1,-1
        call pseudo(q2,q2,expdwb)
        phib = phib + q1(:,:,:,n)*q2
    end do
!   diffusion of the AB junction segment
    call pseudo(q2,q2,expdwb)
!   calculates the total partition function (at the AB junction)
    QQ = dot(q1(:,:,:,NNf),q2)
!   the junction is half A and half B
    phia = 0.5d0*q1(:,:,:,NNF)*q2
    phib = phib + phia
!
!   diffusion of q2 in A part
    do n=NNf-1,1,-1
        call pseudo(q2,q2,expdwa)
        phia = phia + q1(:,:,:,n)*q2
    end do
!   last segment diffusion (A end)
    call pseudo(q2,q2,expdwa)
!   only half contribution from the end
    phia = phia + 0.5d0*q2
!   normalize the concentration
    phia = phia*volume/QQ/NN
    phib = phib*volume/QQ/NN
!   check the mass conservation (if the program is working correctly, temp = 1.0)
    temp = sum(seg*(phia+phib))/volume
    write(*,'(D13.3,D18.8)') temp-1.0d0, QQ
!
end subroutine findphi
!--------------------------------------------------------------------
! this subroutine integrates the block copolymer partition function, q,
! one time-step, ds, forward where expdw = exp(-w*ds/2) and w is the field.
subroutine pseudo(qin,qout,expdw)

    use MKL_DFTI
    use parameters
    implicit none
    real(kind=8), intent(in) :: qin(1:II,1:JJ,1:KK)
    real(kind=8), intent(out) :: qout(1:II,1:JJ,1:KK)
    real(kind=8), intent(in) :: expdw(1:II,1:JJ,1:KK)
    complex(kind=8) :: k_qin(0:II/2, 0:JJ-1, 0:KK-1)
    integer :: i,j,k

! evaluate e^(-w*ds/2) in real space
    do i=1,II
      do j=1,JJ
        do k=1,KK
          qout(i,j,k) = expdw(i,j,k)*qin(i,j,k)
        end do
      end do
    end do

! 3D fourier discrete transform, forward
    call dft_r2c_3d(qout, k_qin)

! multiply e^(-k^2 ds/6) in fourier space, in all 3 directions
    do i=0,II/2
      do j=0,JJ-1
        do k=0,KK-1
          k_qin(i,j,k) = expfactor(i,j,k)*k_qin(i,j,k)
        end do
      end do
    end do

! 3D fourier discrete transform, backword
    call dft_c2r_3d(k_qin, qout)

! normalization calculation
! evaluate e^(-w*ds/2) in real space
    do i=1,II
      do j=1,JJ
        do k=1,KK
          qout(i,j,k) = expdw(i,j,k)*qout(i,j,k)/(II*JJ*KK)
        end do
      end do
    end do

end subroutine pseudo
!--------------------------------------------------------------------
subroutine dft_r2c_3d(rdata, cdata)
    ! compute 3D forward fourier transform

    use MKL_DFTI
    use parameters
    implicit none
    real(kind=8), intent(in) :: rdata(1:II,1:JJ,1:KK)
    complex(kind=8), intent(inout) :: cdata(0:II/2, 0:JJ-1, 0:KK-1)

    ! Execution status
    integer :: status = 0, ignored_status

    ! Strides define data layout for real and complex domain
    integer cstrides(4), rstrides(4)

    ! DFTI descriptor handle
    type(DFTI_DESCRIPTOR), POINTER :: hand
    hand => null()
    cstrides = [0, 1, II/2+1, (int(II/2.0d0)+1)*JJ]
    rstrides = [0, 1, II,     II*JJ]

    status = DftiCreateDescriptor(hand, DFTI_DOUBLE, DFTI_REAL, 3, [II,JJ,KK])
    status = DftiSetValue(hand, DFTI_PLACEMENT, DFTI_NOT_INPLACE)
    status = DftiSetValue(hand, DFTI_CONJUGATE_EVEN_STORAGE, DFTI_COMPLEX_COMPLEX)

    status = DftiSetValue(hand, DFTI_INPUT_STRIDES, rstrides)
    status = DftiSetValue(hand, DFTI_OUTPUT_STRIDES, cstrides)

    status = DftiCommitDescriptor(hand)
    status = DftiComputeForward(hand, rdata(:,1,1), cdata(:,0,0))

    ignored_status = DftiFreeDescriptor(hand)
    end subroutine
!--------------------------------------------------------------------
subroutine dft_c2r_3d(cdata, rdata)
! compute 3D forward fourier transform

use MKL_DFTI
use parameters
implicit none
complex(kind=8), intent(in) :: cdata(0:II/2, 0:JJ-1, 0:KK-1)
real(kind=8), intent(inout) :: rdata(1:II,1:JJ,1:KK)

! Execution status
integer :: status = 0, ignored_status

! Strides define data layout for real and complex domain
integer cstrides(4), rstrides(4)

! DFTI descriptor handle
type(DFTI_DESCRIPTOR), POINTER :: hand
hand => null()
cstrides = [0, 1, II/2+1, (int(II/2.0d0)+1)*JJ]
rstrides = [0, 1, II,     II*JJ]

status = DftiCreateDescriptor(hand, DFTI_DOUBLE, DFTI_REAL, 3, [II,JJ,KK])
status = DftiSetValue(hand, DFTI_PLACEMENT, DFTI_NOT_INPLACE)
status = DftiSetValue(hand, DFTI_CONJUGATE_EVEN_STORAGE, DFTI_COMPLEX_COMPLEX)

status = DftiSetValue(hand, DFTI_INPUT_STRIDES, cstrides)
status = DftiSetValue(hand, DFTI_OUTPUT_STRIDES, rstrides)

status = DftiCommitDescriptor(hand)
status = DftiComputeBackward(hand, cdata(:,0,0), rdata(:,1,1))

ignored_status = DftiFreeDescriptor(hand)
end subroutine
!--------------------------------------------------------------------
