!-------------------------------------------------------------------
!
!   this program solves AB block copolymer melt BCC SPH structure.
!   3-dimensional SCFT (real space method + Neumann boundary +
!   operator splitting + Douglas-Gunn ADI + Anderson mixing) is chosen.
!
!-------------------------------------------------------------------
!this module defines parameters defined once and used repeatedly.
!only static parameters are included in this module
module parameters
    implicit none
!   chiN = interaction parameter between A and B monomers,  pi = 3.141592..
    double precision :: chiN
    double precision, parameter :: pi = atan(1.0d0)*4.0d0
!
!   grid parameters
!   II, JJ, KK = number of grids in x, y and z direction
!   NN = number of time steps (in s direction)
!   NNf = number of time steps for A fractio
    integer :: II, JJ, KK, NN, NNf
!   Lx, Ly, Lz = length of the system in each direction (in units of aN^1/2)
!   dx, dy, dz, ds = discrete step sizes
!   f = A fraction (1-f is the B fraction)
!   alphax, alphay, alphaz = numerical factors repeatedly used
    double precision :: Lx, Ly, Lz, dx, dy, dz, ds, f, alphax, alphay, alphaz
!   xdiag, ydiag, zdiag : 1d array of diagonal elements of
!   the tridiagonal matrix for the ADI method
!   xleft, yleft, zleft : 1d array of the elements left to the diagonal ones
!   xright, yright, zright : 1d array of the elements right to the diagonal ones
!   seg = simple integral weight (half contribution at both edges)
    double precision, allocatable :: xdiag(:),xleft(:),xright(:)
    double precision, allocatable :: ydiag(:),yleft(:),yright(:)
    double precision, allocatable :: zdiag(:),zleft(:),zright(:)
    double precision, allocatable :: segx(:),segy(:),segz(:),seg(:,:,:)
!   volume = volume of the system
    double precision :: volume
end module parameters
!
!--------------------------------------------------------------------
!   The main program begins here
program three_dim_block_copolymer
!
    use parameters
    implicit none
!   iter = number of iteration steps, maxiter = maximum number of iteration steps
    integer :: i, j, k, iter, maxiter
    logical :: has_input
!   parameters related to Anderson mixing
    integer :: nit, nit2, n_anderson, start_anderson, max_anderson, m
!   QQ = total partition function
    double precision :: QQ, energy_chain, energy_field, energy_tot, energy_old
!   error_level = variable to check convergence of the field
!   dot = function performing dot product
!   multidot = function performing multiple dot product
!   mix = dynamically changing mixing parameter
!   min_mix = its minimum value
    double precision :: error_level, old_error_level, dot, multidot
    double precision :: mix, min_mix, temp
!   input(w) and output(wout) fields, xi is temporarily accepted pressures
    double precision, allocatable :: w(:,:,:,:),wout(:,:,:,:), xi(:,:,:)
!   previous field values stored for Anderson mixing
    double precision, allocatable :: wout_hist(:,:,:,:,:),wdiff_hist(:,:,:,:,:)
!   arrays to calculate Anderson mixing
    double precision, allocatable :: u_nm(:,:), v_n(:), a_n(:), wdiffdots(:,:)
!   A and B segment concentration and total segment concentration
    double precision, allocatable :: phia(:,:,:), phib(:,:,:), phitot(:,:,:)
!   strings to name output files
    character (len=5) extension
    character (len=20) label
!
!   has_input = .true. if input file is used
    has_input = .false.
!
    if (has_input) then
        open(10,file='input.txt', status='unknown')
!       read all the stored parameters
        read(10,*) extension, mix, chiN, f, energy_tot, error_level
        read(10,*) NN, II, JJ, KK, Lx, Ly, Lz
!       read fields and concentrations from the file
        allocate(w(1:2,0:II,0:JJ,0:KK),wout(1:2,0:II,0:JJ,0:KK),xi(0:II,0:JJ,0:KK))
        allocate(phia(0:II,0:JJ,0:KK),phib(0:II,0:JJ,0:KK),phitot(0:II,0:JJ,0:KK))
        do i=0,II
            do j=0,JJ
                do k=0,KK
                    read(10,*) w(1,i,j,k), phia(i,j,k), w(2,i,j,k), phib(i,j,k)
                end do
            end do
        end do
        close(10)
        phitot = phia + phib
!
    else
!   if has_input = .false. set up the parameters of the system we solve
!   block copolymer sepcification
        chiN = 18.0d0
        f = 0.25d0
        II = 32
        Lx = 0.93d0
        JJ = 32
        Ly = 0.93d0
        KK = 32
        Lz = 0.93d0
        NN = 100
    end if
!
!   mixing parameter
    mix = 0.2d0
!   minimum mixing parameter
    min_mix = 0.1d0
!   number of iterations before Anderson mixing begins.
!   for now, we assign a large value. it will be determined later.
    start_anderson = 10000
!   max number of previous steps to calculate new field when using Anderson mixing
    max_anderson = 30
    m = max_anderson + 1
!   number of Anderson mixing steps, increases from 1 to max_anderson
    n_anderson = 0
!   iteration must stop before iter = maxiter happens
    maxiter = 1000
!
!   parameter initialization
    call initialize()
!
!   define array sizes for Anderson mixing
    allocate(wout_hist(0:max_anderson,1:2,0:II,0:JJ,0:KK),wdiff_hist(0:max_anderson,1:2,0:II,0:JJ,0:KK))
    allocate(u_nm(1:max_anderson,1:max_anderson),v_n(1:max_anderson),a_n(1:max_anderson),wdiffdots(0:max_anderson,0:max_anderson))
!
!   create initial field
    if(.not. has_input) then
!   define the array sizes of field and concentration
        allocate(w(1:2,0:II,0:JJ,0:KK),wout(1:2,0:II,0:JJ,0:KK),xi(0:II,0:JJ,0:KK))
        allocate(phia(0:II,0:JJ,0:KK),phib(0:II,0:JJ,0:KK),phitot(0:II,0:JJ,0:KK))
        do i=0,II
            do j=0,JJ
                do k=0,JJ
!   BCC sphere input
                    if( ((i*dx)**2+(j*dy)**2+(k*dz)**2) < 0.4d0**2 .or.&
                        (((II-i)*dx)**2+((JJ-j)*dy)**2+((KK-k)*dz)**2) < 0.4d0**2 ) then
                        phia(i,j,k) = 1.0d0
                    else
                        phia(i,j,k) = 0.0d0
                    end if
                end do
            end do
        end do
        phib = 1.0d0 - phia
        w(1,:,:,:) = chiN*phib
        w(2,:,:,:) = chiN*phia
!   keep the level of field value (in order to prevent drifting)
        call zerofield(w)
    end if
!   assign large initial value for the energy and error
    energy_tot =100.0d0
    error_level = 100.0d0
!
!   iteration begins here
    do iter=1,maxiter
!       for the given fields w_a and w_b, find the segment concentration and Q
        call findphi(exp(-w(1,:,:,:)*ds*0.5d0),exp(-(w(2,:,:,:))*ds*0.5d0),phia,phib,QQ)
        phitot = phia + phib
!       calculate individual energy contribution
        energy_chain = -log(QQ/volume)
        energy_field = (chiN*dot(phia,phib) - dot(w(1,:,:,:),phia) - dot(w(2,:,:,:),phib))/volume
!       calculate the total energy per chain
        energy_old = energy_tot
        energy_tot = energy_chain + energy_field
!       calculate pressure field for the new field calculation
        xi = 0.5d0*(w(1,:,:,:)+w(2,:,:,:)-chiN)
!       calculate output fields
        wout(1,:,:,:) = chiN*phib + xi
        wout(2,:,:,:) = chiN*phia + xi
        call zerofield(wout)
!       error_level measures the "relative distance" between the input and output
!       fields
        old_error_level = error_level
        error_level = sqrt(multidot(wout-w,wout-w)/multidot(w,w))
!       print iteration # and error levels
        write(*,'(I6,F8.3,2F13.9)') iter, mix, energy_tot, error_level
!       check the condition to start Anderson mixing
        if(error_level < 5.0d-2) start_anderson = min(start_anderson,iter)
!       check the condition to end the iteration
        if(error_level < 1.0d-7) exit
!       check the condition to apply the simple mixing method
        if(iter <= start_anderson) then
!           just before the Anderson mixing start accumulating past fields
            if(iter == start_anderson) then
                wout_hist(mod(iter,m),:,:,:,:) = wout
                wdiff_hist(mod(iter,m),:,:,:,:) = wout - w
                wdiffdots(mod(iter,m),mod(iter,m)) &
                    = multidot(wdiff_hist(mod(iter,m),:,:,:,:),wdiff_hist(mod(iter,m),:,:,:,:))
            end if
!           dynamically change mixing parameter
            if(old_error_level < error_level) then
                mix = max(mix*0.7d0,min_mix)
            else
                mix = mix*1.01d0
            end if
!           calculate the new field using simple mixing
            w = (1.0d0-mix)*w + mix*wout
!
        else
!           Anderson mixing begins here
!           number of previous steps to use for Anderson mixing
            n_anderson = min(max_anderson,iter-start_anderson)
!           store the input and output field (the memory is used in a periodic way)
            wout_hist(mod(iter,m),:,:,:,:) = wout
            wdiff_hist(mod(iter,m),:,:,:,:) = wout - w
!           evaluate wdiff dot products for calculating Unm and Vn in Thompson's paper
            wdiffdots(mod(iter,m),mod(iter,m)) &
                = multidot(wdiff_hist(mod(iter,m),:,:,:,:),wdiff_hist(mod(iter,m),:,:,:,:))
            do nit=1,n_anderson
                wdiffdots(mod(iter,m),mod(iter-nit,m)) &
                    = multidot(wdiff_hist(mod(iter,m),:,:,:,:),wdiff_hist(mod(iter-nit,m),:,:,:,:))
                wdiffdots(mod(iter-nit,m),mod(iter,m)) = wdiffdots(mod(iter,m),mod(iter-nit,m))
            end do
!           calculate Unm and Vn
            do nit=1,n_anderson
                v_n(nit) =  wdiffdots(mod(iter,m),mod(iter,m)) &
                    - wdiffdots(mod(iter,m),mod(iter-nit,m))
                do nit2=1,n_anderson
                    u_nm(nit,nit2) =  wdiffdots(mod(iter,m),mod(iter,m)) &
                        - wdiffdots(mod(iter,m),mod(iter-nit,m))&
                        - wdiffdots(mod(iter-nit2,m),mod(iter,m))&
                        + wdiffdots(mod(iter-nit,m),mod(iter-nit2,m))
                end do
            end do
!
            call find_an(u_nm(1:n_anderson,1:n_anderson),v_n(1:n_anderson),a_n(1:n_anderson),n_anderson)
!           calculate the new field using Anderson mixing
            w = wout_hist(mod(iter,m),:,:,:,:)
            do nit=1,n_anderson
                w = w + a_n(nit)*(wout_hist(mod(iter-nit,m),:,:,:,:)-wout_hist(mod(iter,m),:,:,:,:))
            end do
        end if
!
!       create intermediate output file at every 100 steps
        if(mod(iter,100) == 0) then
            write (extension,'(I0.4)') iter
            label = 'output.' // extension
!
            open(20,file=label,status='unknown')
            write(20,'(I8,3F8.3,2F13.9)') iter, mix, chiN, f, energy_tot, error_level
            write(20,'(4I5,3F8.3)') NN, II, JJ, KK, Lx, Ly, Lz
            write(20,*) " "
            do i=0,II
                do j=0,JJ
                    do k=0,KK
                        write(20,'(4F12.6)') w(1,i,j,k), phia(i,j,k), w(2,i,j,k), phib(i,j,k)
                    end do
                end do
            end do
            close(20)
        end if
!   the main loop ends here
    end do
!
!   write the final output
    open(30,file='output.txt',status='unknown')
    write(30,'(I8,3F8.3,2F13.9)') iter, mix, chiN, f, energy_tot, error_level
    write(30,'(4I5,3F8.3)') NN, II, JJ, KK, Lx, Ly, Lz
    write(30,*) " "
    do i=0,II
        do j=0,JJ
            do k=0,KK
                write(30,'(4F12.6)') w(1,i,j,k), phia(i,j,k), w(2,i,j,k), phib(i,j,k)
            end do
        end do
    end do
    close(30)
!
end program three_dim_block_copolymer
!
!--------------------------------------------------------------------
!this subroutine initializes repeatedly used parameters
subroutine initialize()
!
    use parameters
    implicit none
    integer :: i, j, k
!
!   s step number for A fraction
    NNf = NN*f
    if( abs(NNf - NN*f) > 1d-8) then
        write(*,*) "NN*f is not an integer."
        stop
    end if
!   grid sizes in x, y, z and s direction
    dx = Lx/II
    dy = Ly/JJ
    dz = Lz/KK
    ds = 1.0d0/NN
    alphax = ds/dx**2/6.0d0
    alphay = ds/dy**2/6.0d0
    alphaz = ds/dz**2/6.0d0
    volume = Lx*Ly*Lz
!   define matrix size
    allocate(xdiag(0:II),xleft(0:II),xright(0:II))
    allocate(ydiag(0:JJ),yleft(0:JJ),yright(0:JJ))
    allocate(zdiag(0:KK),zleft(0:KK),zright(0:KK))
    allocate(seg(0:II,0:JJ,0:KK),segx(0:II),segy(0:JJ),segz(0:KK))
!
!   initialize matrix elements for the diffusion equation solver
!   Neumann boundaries are currently used
    xleft(0) = 0.0d0
    xdiag(0) = 1.0d0+alphax
    xright(0) = -alphax
    do i=1,II-1
        xleft(i) = -0.5d0*alphax
        xdiag(i) = 1.0d0+alphax
        xright(i) = -0.5d0*alphax
    end do
    xleft(II) = -alphax
    xdiag(II) = 1.0d0+alphax
    xright(II) = 0.0d0
!
    yleft(0) = 0.0d0
    ydiag(0) = 1.0d0+alphay
    yright(0) = -alphay
    do j=1,JJ-1
        yleft(j) = -0.5d0*alphay
        ydiag(j) = 1.0d0+alphay
        yright(j) = -0.5d0*alphay
    end do
    yleft(JJ) = -alphay
    ydiag(JJ) = 1.0d0+alphay
    yright(JJ) = 0.0d0
!
    zleft(0) = 0.0d0
    zdiag(0) = 1.0d0+alphaz
    zright(0) = -alphaz
    do k=1,KK-1
        zleft(k) = -0.5d0*alphaz
        zdiag(k) = 1.0d0+alphaz
        zright(k) = -0.5d0*alphaz
    end do
    zleft(KK) = -alphaz
    zdiag(KK) = 1.0d0+alphaz
    zright(KK) = 0.0d0
!   weight factor for three-dim integral
!   only half contribution at the boundary(trapezoid integral)
    segx(:) = 1.0d0 * dx
    segx(0) = 0.5d0 * dx
    segx(II) = 0.5d0 * dx
    segy(:) = 1.0d0 * dy
    segy(0) = 0.5d0 * dy
    segy(JJ) = 0.5d0 * dy
    segz(:) = 1.0d0 * dz
    segz(0) = 0.5d0 * dz
    segz(KK) = 0.5d0 * dz
!
    do i=0,II
        do j=0,JJ
            do k=0,KK
                seg(i,j,k) = segx(i)*segy(j)*segz(k)
            end do
        end do
    end do
!
end subroutine initialize
!
!--------------------------------------------------------------------
!This subroutine keeps the level of field
subroutine zerofield(w)
!
    use parameters
    implicit none
    double precision, intent(inout) :: w(1:2,0:II,0:JJ,0:KK)
    double precision :: tot1, tot2, temp
!   each sum of w_a and w_b is set to zero
    tot1 = sum(seg(:,:,:)*w(1,:,:,:))
    w(1,:,:,:) = w(1,:,:,:) - tot1/volume
    tot2 = sum(seg(:,:,:)*w(2,:,:,:))
    w(2,:,:,:) = w(2,:,:,:) - tot2/volume
!
end subroutine zerofield
!
!--------------------------------------------------------------------
!this function calculates three-dim integral g*h
double precision function dot(g,h)
!
    use parameters
    implicit none
    double precision, intent(in) :: g(0:II,0:JJ,0:KK), h(0:II,0:JJ,0:KK)
    dot = sum(seg(:,:,:)*g(:,:,:)*h(:,:,:))
end function dot
!
!--------------------------------------------------------------------
!this function calculates multiple three-dim integrals
double precision function multidot(g,h)
!
    use parameters
    implicit none
    double precision, intent(in) :: g(1:2,0:II,0:JJ,0:KK), h(1:2,0:II,0:JJ,0:KK)
    multidot = sum(seg(:,:,:)*(g(1,:,:,:)*h(1,:,:,:)+g(2,:,:,:)*h(2,:,:,:)))
end function multidot
!
!--------------------------------------------------------------------
!this subroutine finds the solution for the matrix equation (n*n size)
!for Anderson mixing U * A = V
!Gauss elimination method is used in its simplest level
subroutine find_an(u,v,a,n)
    implicit none
    integer, intent(in) :: n
    double precision, intent(inout) :: u(1:n,1:n), v(1:n)
    double precision, intent(out) :: a(1:n)
!
    integer :: i,j,k
    double precision :: factor, tempsum
!   Elimination process
    do i=1,n
        do j=i+1,n
            factor = u(j,i)/u(i,i)
            v(j) = v(j) - v(i)*factor
            do k=i+1,n
                u(j,k) = u(j,k) - u(i,k)*factor
            end do
        end do
    end do
!
!   find the solution
    a(n) = v(n)/u(n,n)
    do i=n-1,1,-1
        tempsum = 0.0d0
        do j=i+1,n
            tempsum = tempsum + u(i,j)*a(j)
        end do
        a(i) = (v(i) - tempsum)/u(i,i)
    end do
!
end subroutine find_an
!
!--------------------------------------------------------------------
!this subroutine calculates the segment concentration
!where expdwa = exp(-wa*ds/2) and expdwb = exp(-wb*ds/2)
subroutine findphi(expdwa,expdwb,phia,phib,QQ)
!
    use parameters
    implicit none
    double precision, intent(in) :: expdwa(0:II,0:JJ,0:KK), expdwb(0:II,0:JJ,0:KK)
    double precision, intent(out) :: phia(0:II,0:JJ,0:KK), phib(0:II,0:JJ,0:KK)
    double precision, intent(out) :: QQ
    double precision :: q1(0:II,0:JJ,0:KK,0:NN), q2(0:II,0:JJ,0:KK)
    double precision :: dot, temp
    integer :: n, i, j, k
!
!   q1 is q and q2 is qdagger in the note
!   free end initial condition (q1 starts from A end)
    q1(:,:,:,0) = 1.0d0
!
!   diffusion of A chain
    do n=1,NNf
        call crank(q1(:,:,:,n-1),q1(:,:,:,n),expdwa)
    end do
!
!   diffusion of B chain
    do n=NNf+1,NN
        call crank(q1(:,:,:,n-1),q1(:,:,:,n),expdwb)
    end do
!
!   q2 starts from B end
    q2 = 1.0d0
!   segment concentration from the end. (note the half contribution)
    phib = 0.5d0*q1(:,:,:,NN)
!
!   diffusion of q2 in B part
    do n=NN-1,NNf+1,-1
        call crank(q2,q2,expdwb)
        phib = phib + q1(:,:,:,n)*q2
    end do
!   diffusion of the AB junction segment
    call crank(q2,q2,expdwb)
!   calculates the total partition function (at the AB junction)
    QQ = dot(q1(:,:,:,NNf),q2)
!   the junction is half A and half B
    phia = 0.5d0*q1(:,:,:,NNF)*q2
    phib = phib + phia
!
!   diffusion of q2 in A part
    do n=NNf-1,1,-1
        call crank(q2,q2,expdwa)
        phia = phia + q1(:,:,:,n)*q2
    end do
!   last segment diffusion (A end)
    call crank(q2,q2,expdwa)
!   only half contribution from the end
    phia = phia + 0.5d0*q2
!   normalize the concentration
    phia = phia*volume/QQ/NN
    phib = phib*volume/QQ/NN
!   check the mass conservation (if the program is working correctly, temp = 1.0)
    temp = sum(seg*(phia+phib))/volume
    write(*,'(D13.3,D18.8)') temp-1.0d0, QQ
!
end subroutine findphi
!
!--------------------------------------------------------------------
!this subroutine integrates the block copolymer partition function, q
!one time-step, ds, forward where expdw = exp(-w*ds/2) and w is the field
!ADI in Zeng and Zhu's paper is used
subroutine crank(qin,qout,expdw)
!
    use parameters
    implicit none
    double precision, intent(in) :: qin(0:II,0:JJ,0:KK)
    double precision, intent(out) :: qout(0:II,0:JJ,0:KK)
    double precision, intent(in) :: expdw(0:II,0:JJ,0:KK)
    double precision, dimension(0:II,0:JJ,0:KK) :: qa, qb, qstar, qstar2
    double precision :: temp1(0:II),temp2(0:JJ),temp3(0:KK)
    integer :: i, im, ip, j, jm, jp, k, km, kp
!
!   the first of the three step evaluation
    qa= expdw*qin
!
!   the first equation
    do k=0,KK
        km = max(0,k-1)
        kp = min(KK,k+1)
        do j=0,JJ
            jm = max(0,j-1)
            jp = min(JJ,j+1)
            do i=0,II
                im = max(0,i-1)
                ip = min(II,i+1)
!   explicit calculation of the B part of Ax=B matrix equation
                temp1(i) = 2.0d0*((3.0d0-0.5d0*xdiag(i)-ydiag(j)-zdiag(k))*qa(i,j,k) &
                    - yleft(j)*qa(i,jm,k) - yright(j)*qa(i,jp,k) &
                    - zleft(k)*qa(i,j,km) - zright(k)*qa(i,j,kp)) &
                    - xleft(i)*qa(im,j,k) - xright(i)*qa(ip,j,k)
            end do
!   solve the matrix equation for the implicit part of the ADI method
            call tridiagonal(xleft,xdiag,xright,qstar(:,j,k),temp1,0,II)
        end do
    end do
!
!   second equation
    do k=0,KK
        do i=0,II
            do j=0,JJ
                jm = max(0,j-1)
                jp = min(JJ,j+1)
    !   explicit calculation of the B part of Ax=B matrix equation
                temp2(j) = qstar(i,j,k) + (ydiag(j)-1.0d0)*qa(i,j,k) + yleft(j)*qa(i,jm,k) + yright(j)*qa(i,jp,k)
            end do
    !   solve the matrix equation for the implicit part of the ADI method
            call tridiagonal(yleft,ydiag,yright,qstar2(i,:,k),temp2,0,JJ)
        end do
    end do
!   third equation
    do j=0,JJ
        do i=0,II
            do k=0,KK
                km = max(0,k-1)
                kp = min(KK,k+1)
  !   explicit calculation of the B part of Ax=B matrix equation
                temp3(k) = qstar2(i,j,k) + (zdiag(k)-1.0d0)*qa(i,j,k) + zleft(k)*qa(i,j,km) + zright(k)*qa(i,j,kp)
            end do
  !   solve the matrix equation for the implicit part of the ADI method
            call tridiagonal(zleft,zdiag,zright,qb(i,j,:),temp3,0,KK)
        end do
    end do
!
!   the third step
    qout = expdw*qb

end subroutine crank
!
!--------------------------------------------------------------------
!this subroutine solves CX=Y, where C is a tridiagonal matrix
!cu, cd, cl supplies the tridiagonal part of the matrix
!the starting index of the array is n0 (either 0 or 1)
subroutine tridiagonal(cu,cd,cl,x,y,n0,nm)
!
    implicit none
    integer, intent(in) :: n0, nm
    double precision, intent(in) :: cu(0:nm), cd(0:nm), cl(0:nm)
    double precision, intent(out) :: x(0:nm)
    double precision, intent(in) :: y(0:nm)
    double precision :: gam(0:nm), bet
    integer :: j
!
    x(0) = 0.0d0
!   if (cd(n0).eq.0.0d0) stop 'Zero matrix element'
    bet = cd(n0)
    x(n0) = y(n0)/bet
    do j=1,nm
        gam(j) = cl(j-1)/bet
        bet = cd(j)-cu(j)*gam(j)
!       if (bet.eq.0.0d0) stop 'Bad zero term in tridiagonal'
        x(j) = (y(j)-cu(j)*x(j-1))/bet
    end do
    do j=nm-1,n0,-1
        x(j) = x(j) - gam(j+1)*x(j+1)
    end do
end subroutine tridiagonal
