!-------------------------------------------------------------------
!
!   this program solves 1-dimensional polymer brush in good solvents
!   a simple 1-dimensional SCFT (real space method + Dirichlet boundary +
!   operator splitting + Crank-Nicolson) is chosen.
!
!--------------------------------------------------------------------
!this module defines parameters defined once and used repeatedly.
!only static parameters are included in this module
module parameters
    implicit none
!   l0 = brush height in strong stretching theory,
!   Lambda = interaction parameter, pi = 3.141592..
    double precision :: l0, Lambda
    double precision, parameter :: pi = atan(1.0d0)*4.0d0
!
!   grid parameters
!   KK = number of grids in z direction
!   NN = number of time steps (in s direction)
    integer :: KK, NN
!   Lz = length of the system (in units of aN^1/2)
!   dz, ds = discrete step sizes
!   alphaz = numerical factor repeatedly used
    double precision :: Lz, dz, ds, alphaz
!   zdiag : 1d array of diagonal elements of the tridiagonal matrix for Crank-Nicolson method
!   zleft : 1d array of the elements left to the diagonal ones
!   zright : 1d array of the elements right to the diagonal ones
!   seg = simple integral weight (half contribution at both edges)
    double precision, allocatable :: zdiag(:), zleft(:), zright(:), seg(:)
end module parameters
!
!--------------------------------------------------------------------
!   The main program begins here
program one_dim_brush_solvent
!
    use parameters
    implicit none
!   iter = number of iteration steps, maxiter = maximum number of iteration steps
    integer :: k, iter, maxiter
!   qq = total partition function
    double precision :: qq, energy_chain, energy_field, energy_tot, energy_old
!   error_level = variable to check convergence of the field
!   dot = function performing dot product
!   mix = dynamically changing mixing parameter
!   min_mix = its minimum value
    double precision :: error_level, old_error_level, dot
    double precision :: mix, min_mix, temp
!   input(w) and output(woout) fields
    double precision, allocatable :: w(:), wout(:)
!   segment concentration(polymer density)
    double precision, allocatable :: phi(:)
!
!   set up the parameters of the system we solve
!   brush sepcification
    l0 = 2.0d0
    KK = 250
    Lz = 5.0d0
    NN = 1500
!
!   mixing parameter
    mix = 0.2d0
!   minimum mixing parameter
    min_mix = 0.1d0
!   iteration must stop before iter = maxiter happens
    maxiter = 1000
!
!   parameter initialization
    call initialize()
!
!   define the array sizes of field and concentration
    allocate(w(0:KK),wout(0:KK),phi(0:KK))
!   initialize the trial field
    do k=1,KK-1
        w(k) = Lambda*1.5d0*(1.0d0-(k*dz/l0)**2) !   use SST prediction as an input
        if(w(k)<0.0d0) w(k) = 0.0d0
    end do
!   assign large initial value for the energy and error
    energy_tot =100.0d0
    error_level = 100.0d0
!
!   iteration begins here
    do iter=1, maxiter
!       for the given fields w, find the segment concentration and Q
        call find_phi(exp(-w(:)*ds*0.5d0),phi,qq)
!       calculate individual energy contribution
        energy_chain = -log(qq)
        energy_field = dot(w(:),phi(:))
!       calculate the total energy per chain (field energy is double counted, so
!       substract half)
        energy_old = energy_tot
        energy_tot = energy_chain - energy_field/2.0d0
!       calculate output field
        wout(:) = Lambda*phi(:)
!       error_level measures the "relative distance" between the input and output fields
        old_error_level = error_level
        error_level = sqrt(dot(wout-w,wout-w)/dot(w,w))
!       print iteration # and error levels
        write(*,'(I6,F8.3,2F13.9)') iter, mix, energy_tot, error_level
!       check the condition to end the iteration
        if(error_level < 1.0d-7) exit
!       dynamically change mixing parameter
        if(old_error_level < error_level) then
            mix = max(mix*0.7d0,min_mix)
        else
            mix = mix*1.01d0
        end if
!       calculate the new field using simple mixing
        w = (1.0d0-mix)*w + mix*wout
!
!   the main loop ends here
    end do
!
!   write the final output
    open(30,file='output.txt',status='unknown')
    write(30,'(I8,2F8.3,2F13.9)') iter, mix, l0, energy_tot, error_level
    write(30,'(2I5,F8.3)') NN, KK, Lz
    write(30,*) " "
    do k=0,KK
        write(30,'(2F12.6)') w(k), phi(k)
    end do
    close(30)
!
end program one_dim_brush_solvent
!
!--------------------------------------------------------------------
!this subroutine initializes repeatedly used parameters
subroutine initialize()
!
    use parameters
    implicit none
    integer :: k
!
!   relation between Lambda and l0
    Lambda = (l0**3)*pi*pi/4.0d0
!   grid sizes in z and s direction
    dz = Lz/KK
    ds = 1.0d0/NN
    alphaz = ds/dz**2/6.0d0
!   define matrix size
    allocate(zdiag(0:KK),zleft(0:KK),zright(0:KK),seg(0:KK))
!
!   initialize matrix elements for the diffusion equation solver
!   Dirichlet boundaries are currently used (0th and KK'th elements are not
!   defined)
    zleft(1) = 0.0d0
    zdiag(1) = 1.0d0+alphaz
    zright(1) = -0.5d0*alphaz
    do k=2,KK-2
        zleft(k) = -0.5d0*alphaz
        zdiag(k) = 1.0d0+alphaz
        zright(k) = -0.5d0*alphaz
    end do
    zleft(KK-1) = -0.5d0*alphaz
    zdiag(KK-1) = 1.0d0+alphaz
    zright(KK-1) = 0.0d0
!   weight factor for integral in z direction
!   only half contribution at the boundary(trapezoid integral)
    seg(:) = 1.0d0 * dz
    seg(0) = 0.5d0 * dz
    seg(KK) = 0.5d0 * dz
!
end subroutine initialize
!
!--------------------------------------------------------------------
!this function calculates integral g*h in z direction
double precision function dot(g,h)
!
    use parameters
    implicit none
    double precision, intent(in) :: g(0:KK), h(0:KK)
    dot = sum(seg(:)*g(:)*h(:))
end function dot
!
!--------------------------------------------------------------------
!this subroutine calculates the segment concentration phi
!where expdw = exp(-w*ds/2)
subroutine find_phi(expdw,phi,QQ)
!
    use parameters
    implicit none
    double precision, intent(in) :: expdw(0:KK)
    double precision, intent(out) :: phi(0:KK)
    double precision, intent(out) :: QQ
    double precision :: q1(0:KK,0:NN), q2(0:KK,0:NN)
    double precision :: dot, temp
    integer :: n, k
!
!   q1 is q and q2 is qdagger in the standard SCFT
!   free end initial condition (q1 starts from the free end)
    q1(:,0) = 1.0d0
    q1(0,0) = 0.0d0
    q1(KK,0) = 0.0d0
!
!   diffusion of chain
    do n=1,NN
        call crank(q1(:,n-1),q1(:,n),expdw)
    end do
!
!   q2 starts from the grafted end, delta function initial condition at k=1
    q2(:,NN) = 0.0d0
    q2(1,NN) = 1.0d0/dz
!   diffusion of chain
    do n=NN-1,0,-1
        call crank(q2(:,n+1),q2(:,n),expdw)
    end do
!
!   calculates the total partition function
    QQ = dot(q1(:,0),q2(:,0))
!
!   calculates the segment concentration
    phi(:) = (q1(:,0)*q2(:,0)+q1(:,NN)*q2(:,NN))/2.0d0
    do n=1,NN-1
        phi(:) = phi(:) + q1(:,n)*q2(:,n)
    end do
!   normalize the concentration
    phi = phi/QQ/NN
!
!   check the mass conservation (if the program is working correctly, temp = 1.0)
    temp = sum(seg(:)*(phi(:)))
    write(*,'(D13.3,D18.8)') temp-1.0d0, QQ
!
end subroutine find_phi
!
!--------------------------------------------------------------------
!this subroutine integrates the block copolymer partition function, q
!one time-step, ds, forward where expdw = exp(-w*ds/2) and w is the field
!operator splitted Crank-Nicolson method is used
subroutine crank(qin,qout,expdw)
!
    use parameters
    implicit none
    double precision, intent(in) :: qin(0:KK)
    double precision, intent(out) :: qout(0:KK)
    double precision, intent(in) :: expdw(0:KK)
    double precision :: qstar(0:KK)
    integer :: k, km, kp
!
!   the first of the three step evaluation
    qout = expdw*qin
!
!   the second step
    do k=1,KK-1
        km = k-1
        kp = k+1
!   explicit calculation of the B part of Ax=B matrix equation
        qstar(k) = (2.0d0-zdiag(k))*qout(k) - zleft(k)*qout(km) - zright(k)*qout(kp)
    end do
!   solve the matrix equation for the implicit part of the CN method
    call tridiagonal(zleft,zdiag,zright,qout,qstar,1,KK-1)
!
!   the third step
    qout = expdw*qout
!
end subroutine crank
!
!--------------------------------------------------------------------
!this subroutine solves CX=Y, where C is a tridiagonal matrix
!cu, cd, cl supplies the tridiagonal part of the matrix
!the starting index of the array is n0 (either 0 or 1)
subroutine tridiagonal(cu,cd,cl,x,y,n0,nm)
!
    implicit none
    integer, intent(in) :: n0, nm
    double precision, intent(in) :: cu(0:nm), cd(0:nm), cl(0:nm)
    double precision, intent(out) :: x(0:nm)
    double precision, intent(in) :: y(0:nm)
    double precision :: gam(0:nm), bet
    integer :: j
!
    x(0) = 0.0d0
!   if (cd(n0).eq.0.0d0) stop 'Zero matrix element'
    bet = cd(n0)
    x(n0) = y(n0)/bet
    do j=1,nm
        gam(j) = cl(j-1)/bet
        bet = cd(j)-cu(j)*gam(j)
!       if (bet.eq.0.0d0) stop 'Bad zero term in tridiagonal'
        x(j) = (y(j)-cu(j)*x(j-1))/bet
    end do
    do j=nm-1,n0,-1
        x(j) = x(j) - gam(j+1)*x(j+1)
    end do
end subroutine tridiagonal
