!-------------------------------------------------------------------
!
!   this program calculates the distribution of one end-fixed chain
!   a simple 1-dimensional (real space method + Neumann boundary +
!   operator splitting + Crank-Nicolson) is chosen.
!
!--------------------------------------------------------------------
!this module defines parameters defined once and used repeatedly.
!only static parameters are included in this module
module parameters
    implicit none
!   pi = 3.141592..
    double precision, parameter :: pi = atan(1.0d0)*4.0d0
!   grid parameters
!   KK = number of grids in z direction
!   NN = number of time steps (in s direction)
    integer :: KK, NN
!   Lz = length of the system (in units of aN^1/2)
!   dz, ds = discrete step sizes
!   alphaz = numerical factor repeatedly used
    double precision :: Lz, dz, ds, alphaz
!   zdiag : 1d array of diagonal elements of the tridiagonal matrix for Crank-Nicolson method
!   zleft : 1d array of the elements left to the diagonal ones
!   zright : 1d array of the elements right to the diagonal ones
!   seg = simple integral weight (half contribution at both edges)
    double precision, allocatable :: zdiag(:), zleft(:), zright(:), seg(:)
end module parameters
!
!--------------------------------------------------------------------
!   The main program begins here
program single_chain_in_field
!
    use parameters
    implicit none
!
    integer :: k
!   qq = total partition function
!   elastic_energy = elastic energy of a chain
    double precision :: qq, elastic_energy
!   dot = function performing dot product
    double precision :: dot, temp
!   potential field w
    double precision, allocatable :: w(:)
!   segment concentration(polymer density)
    double precision, allocatable :: phi(:), phi_end(:), phi_end_prediction(:)
!
!   set up the parameters of the system we solve
    KK = 2000
    Lz = 20.0d0
    NN = 2000
!
!   parameter initialization
    call initialize()
!
!   define the array sizes of field and concentration
    allocate(w(0:KK),phi(0:KK),phi_end(0:KK),phi_end_prediction(0:KK))
!   the input field
    do k=0,KK
        w(k) = 0.0d0
    end do
!
!   for the given fields w, find the segment concentration and Q
    call find_phi(exp(-w(:)*ds*0.5d0),phi,phi_end,qq)
!   calculate individual energy contribution
    elastic_energy = -log(qq) - dot(w(:),phi(:))
!
    do k=0,KK
        phi_end_prediction(k) = (1.5d0/pi)**(0.5d0)*exp(-1.5d0*((k-KK/2)*dz)**2)
    end do
    write(*,*) sum(seg(:)*phi_end_prediction(:)), sum(seg(:)*phi_end(:))

!   write the final output
    open(30,file='output.txt',status='unknown')
    write(30,'(2I5,F8.3,F13.9)') NN, KK, Lz, elastic_energy
    write(30,*) " "
    do k=0,KK
        write(30,'(4F12.6)') w(k), phi(k), phi_end(k), phi_end_prediction(k)
    end do
    close(30)
!
end program single_chain_in_field
!
!--------------------------------------------------------------------
!this subroutine initializes repeatedly used parameters
subroutine initialize()
!
    use parameters
    implicit none
    integer :: k
!
!   grid sizes in z and s direction
    dz = Lz/KK
    ds = 1.0d0/NN
    alphaz = ds/dz**2/6.0d0
!   define matrix size
    allocate(zdiag(0:KK),zleft(0:KK),zright(0:KK),seg(0:KK))
!
!   initialize matrix elements for the diffusion equation solver
!   Neumann boundaries are currently used
    zleft(0) = 0.0d0
    zdiag(0) = 1.0d0+alphaz
    zright(0) = -alphaz
    do k=1,KK-1
        zleft(k) = -0.5d0*alphaz
        zdiag(k) = 1.0d0+alphaz
        zright(k) = -0.5d0*alphaz
    end do
    zleft(KK) = -alphaz
    zdiag(KK) = 1.0d0+alphaz
    zright(KK) = 0.0d0
!   weight factor for integral in z direction
!   only half contribution at the boundary(trapezoid integral)
    seg(:) = 1.0d0 * dz
    seg(0) = 0.5d0 * dz
    seg(KK) = 0.5d0 * dz
!
end subroutine initialize
!
!--------------------------------------------------------------------
!this function calculates integral g*h in z direction
double precision function dot(g,h)
!
    use parameters
    implicit none
    double precision, intent(in) :: g(0:KK), h(0:KK)
    dot = sum(seg(:)*g(:)*h(:))
end function dot
!
!--------------------------------------------------------------------
!   this subroutine calculates the segment concentration phi
!   where expdw = exp(-w*ds/2)
subroutine find_phi(expdw,phi,phi_end,QQ)
!
    use parameters
    implicit none
    double precision, intent(in) :: expdw(0:KK)
    double precision, intent(out) :: phi(0:KK), phi_end(0:KK)
    double precision, intent(out) :: QQ
    double precision :: q1(0:KK,0:NN), q2(0:KK,0:NN)
    double precision :: dot, temp
    integer :: n, k
!
!   q1 is q and q2 is qdagger in the standard SCFT
!   free end initial condition (q1 starts from the free end)
    q1(:,0) = 1.0d0
!
!   diffusion of chain
    do n=1,NN
        call crank(q1(:,n-1),q1(:,n),expdw)
    end do
!
!   the fixed end is at z = Lz/2
    q2(:,NN) = 0.0d0
    q2(KK/2,NN) = 1.0d0/dz
!   diffusion of chain
    do n=NN-1,0,-1
        call crank(q2(:,n+1),q2(:,n),expdw)
    end do
!
!   calculates the total partition function
    QQ = dot(q1(:,0),q2(:,0))
!
!   calculates the segment concentration
    phi(:) = (q1(:,0)*q2(:,0)+q1(:,NN)*q2(:,NN))/2.0d0
    do n=1,NN-1
        phi(:) = phi(:) + q1(:,n)*q2(:,n)
    end do
!   normalize the concentration
    phi = phi/QQ/NN
    phi_end(:) = q1(:,0)*q2(:,0)/QQ
!
!   check the mass conservation (if the program is working correctly, temp = 1.0)
    temp = sum(seg(:)*(phi(:)))
    write(*,'(D13.3,D18.8)') temp-1.0d0, QQ
!
end subroutine find_phi
!
!--------------------------------------------------------------------
!this subroutine integrates the block copolymer partition function, q
!one time-step, ds, forward where expdw = exp(-w*ds/2) and w is the field
!operator splitted Crank-Nicolson method is used
subroutine crank(qin,qout,expdw)
!
    use parameters
    implicit none
    double precision, intent(in) :: qin(0:KK)
    double precision, intent(out) :: qout(0:KK)
    double precision, intent(in) :: expdw(0:KK)
    double precision :: qstar(0:KK)
    integer :: k, km, kp
!
!   the first of the three step evaluation
    qout = expdw*qin
!
!   the second step
    do k=0,KK
        km = max(0,k-1)
        kp = min(KK,k+1)
!   explicit calculation of the B part of Ax=B matrix equation
        qstar(k) = (2.0d0-zdiag(k))*qout(k) - zleft(k)*qout(km) - zright(k)*qout(kp)
    end do
!   solve the matrix equation for the implicit part of the CN method
    call tridiagonal(zleft,zdiag,zright,qout,qstar,0,KK)
!
!   the third step
    qout = expdw*qout
!
end subroutine crank
!
!--------------------------------------------------------------------
!this subroutine solves CX=Y, where C is a tridiagonal matrix
!cu, cd, cl supplies the tridiagonal part of the matrix
!the starting index of the array is n0 (either 0 or 1)
subroutine tridiagonal(cu,cd,cl,x,y,n0,nm)
!
    implicit none
    integer, intent(in) :: n0, nm
    double precision, intent(in) :: cu(0:nm), cd(0:nm), cl(0:nm)
    double precision, intent(out) :: x(0:nm)
    double precision, intent(in) :: y(0:nm)
    double precision :: gam(0:nm), bet
    integer :: j
!
    x(0) = 0.0d0
!   if (cd(n0).eq.0.0d0) stop 'Zero matrix element'
    bet = cd(n0)
    x(n0) = y(n0)/bet
    do j=1,nm
        gam(j) = cl(j-1)/bet
        bet = cd(j)-cu(j)*gam(j)
!       if (bet.eq.0.0d0) stop 'Bad zero term in tridiagonal'
        x(j) = (y(j)-cu(j)*x(j-1))/bet
    end do
    do j=nm-1,n0,-1
        x(j) = x(j) - gam(j+1)*x(j+1)
    end do
end subroutine tridiagonal
